#!/usr/bin/env python
import logging

import click

from kubeyard import logging as kubeyard_logging
from kubeyard import settings
from kubeyard.commands import BuildCommand
from kubeyard.commands import DebugCommand
from kubeyard.commands import DeployCommand
from kubeyard.commands import FixCodeStyleCommand
from kubeyard.commands import InitCommand
from kubeyard.commands import InstallCompletion
from kubeyard.commands import InstallGlobalSecretsCommand
from kubeyard.commands import PushCommand
from kubeyard.commands import SetupCommand
from kubeyard.commands import ShellCommand
from kubeyard.commands import TestCommand
from kubeyard.commands import UpdateRequirementsCommand
from kubeyard.commands.init import PythonPackageInitType
from kubeyard.commands.init import all_templates
from kubeyard.entrypoints.custom_command_loader import CustomCommandsLoader

kubeyard_logging.init_logging()
logger = logging.getLogger(__name__)


@click.group()
@click.version_option()
def cli():
    """
    kubeyard is a commandline tool for easier development and deployment of Social WiFi services.

    If some command allows you to pass additional arguments (for example to pytest),
    you should use double dash `--` to separate it from command args.

    kubeyard [OPTIONS] COMMAND [ARGS] -- [ARGS SHOULD BE PASSED]
    """
    pass


def apply_common_options(options):
    def wrap(func):
        for option in reversed(options):
            func = option(func)
        return func

    return wrap


initialized_repository_options = (
    click.option(
        "--directory",
        default=settings.DEFAULT_KUBEYARD_PROJECT_DIR,
        type=click.Path(
            file_okay=False,
            exists=True,
            resolve_path=True,
        ),
        help="Select project root directory.",
    ),
    click.option(
        "--verbose",
        "-v",
        "log_level",
        flag_value="DEBUG",
        help="Outputs debug logs.",
    ),
)

devel_options = (
    click.option(
        "--tag",
        help="Used image tag.",
    ),
    click.option(
        "--default",
        "use_default_implementation",
        is_flag=True,
        help="Don't try to execute custom script. Useful when you need original behaviour in overridden method.",
    ),
    click.option(
        "--image-name",
        help="Image name (without repository). Default is set in kubeyard.yml.",
    ),
)


@cli.command(
    help=TestCommand.__doc__,
    context_settings=dict(
        ignore_unknown_options=True,
    ),
)
@apply_common_options(initialized_repository_options)
@apply_common_options(devel_options)
@click.option(
    "--force-migrate-db",
    "-f-m-db",
    "force_migrate_database",
    is_flag=True,
    help="On dev environment DB is cached, so if you have some new migrations, you should use this flag.",
)
@click.option(
    "--force-recreate-db",
    "-f-r-db",
    "force_recreate_database",
    is_flag=True,
    help="On dev environment DB is cached, "
         "so you can use this flag to remove existing DB before tests and create new one.",
)
@click.argument("test_options", nargs=-1, type=click.UNPROCESSED)
def test(**kwargs):
    TestCommand(**kwargs).run()


@cli.command(help=FixCodeStyleCommand.__doc__)
@apply_common_options(initialized_repository_options)
@apply_common_options(devel_options)
def fix_code_style(**kwargs):
    FixCodeStyleCommand(**kwargs).run()


@cli.command(help=BuildCommand.__doc__)
@apply_common_options(initialized_repository_options)
@apply_common_options(devel_options)
@click.option(
    "--image-context",
    help="Image context containing Dockerfile. Defaults to <project_dir>/docker",
)
@click.option(
    "--docker-args",
    help='Additional arguments passed to docker command. Use like that: --docker-args="-a x --arg2=y --arg3=z"',
)
def build(**kwargs):
    BuildCommand(**kwargs).run()


@cli.command(help=UpdateRequirementsCommand.__doc__)
@apply_common_options(initialized_repository_options)
@apply_common_options(devel_options)
def update_requirements(**kwargs):
    UpdateRequirementsCommand(**kwargs).run()


@cli.command(help=PushCommand.__doc__)
@apply_common_options(initialized_repository_options)
@apply_common_options(devel_options)
def push(**kwargs):
    PushCommand(**kwargs).run()


@cli.command(help=DeployCommand.__doc__)
@apply_common_options(initialized_repository_options)
@apply_common_options(devel_options)
@click.option(
    "--build-url",
    help="URL to a CI/CD (eg. Jenkins) build. It will be used as a pod annotation.",
)
@click.option(
    "--gcs-service-key-file",
    help="Service account key path for Google Cloud Storage. Required to deploy static files on GCS.",
)
@click.option(
    "--aws-credentials",
    help="AWS credentials for S3 buckets. Required to deploy static files on S3.",
)
@click.option(
    "--azure-connection-string",
    help="Azure connection string for Azure Storage Blob Container. Required to deploy static files on Azure Storage",
)
@click.option(
    "--bucket-name",
    "--gcs-bucket-name",
    "--s3-bucket-name",
    "--azure-container-name",
    help="Google Cloud Storage or S3 bucket name, or Azure container name. Required to deploy static files."
         "May be configured using context variable: 'BUCKET_NAME'",
)
@click.option(
    "--upload-local-binary-path",
    type=click.Path(
        dir_okay=False,
        exists=True,
    ),
    help="Use specified local binary, instead of docker image, for static files uploading.",
)
def deploy(**kwargs):
    DeployCommand(**kwargs).run()


@cli.command(help=DebugCommand.__doc__)
@apply_common_options(initialized_repository_options)
@click.argument(
    'name',
    required=False,
)
def variables(**kwargs):
    DebugCommand(**kwargs).run()


@cli.command(help=InstallCompletion.__doc__)
@click.option(
    "--force",
    "-f",
    is_flag=True,
    help="Force reinstall bash completion.",
)
def install_bash_completion(**kwargs):
    InstallCompletion(**kwargs).run()


@cli.command(help=InstallGlobalSecretsCommand.__doc__)
def install_global_secrets():
    InstallGlobalSecretsCommand().run()


@cli.command(help=SetupCommand.__doc__)
@click.option(
    "--development",
    "mode",
    flag_value="development",
    help="Sets development context using minikube. Development context uses secrets from repository.",
)
@click.option(
    "--production",
    "mode",
    flag_value="production",
)
def setup(**kwargs):
    SetupCommand(**kwargs).run()


@cli.command(help=InitCommand.__doc__)
@click.option(
    "--directory",
    default=".",
    type=click.Path(exists=True, file_okay=False),
    help="Select project root directory.",
)
@click.option(
    "--template",
    "template_name",
    type=click.Choice([template.name for template in all_templates]),
    default=PythonPackageInitType.name,
    help="Select ember template.",
)
def init(*, template_name, **kwargs):
    for template in all_templates:
        if template.name == template_name:
            return InitCommand(init_type=template, **kwargs).run()


@cli.command(help=ShellCommand.__doc__)
@apply_common_options(initialized_repository_options)
@apply_common_options(devel_options)
@click.option(
    "--pod",
    "-p",
    help="You can select pod to exec into.",
)
@click.option(
    "--container",
    "-c",
    help="You can select container to exec into.",
)
@click.option(
    "--shell",
    help="Which shell should be used in exec.",
    default="bash",
)
@click.option(
    "--root",
    is_flag=True,
    help="Do not create user, use root.",
)
def shell(**kwargs):
    ShellCommand(**kwargs).run()


custom_commands = CustomCommandsLoader(cli, help="Collection of all custom commands")

cli_with_custom_commands = click.CommandCollection(sources=[
    cli,
    custom_commands,
])



import logging

import sh

from cached_property import cached_property

from kubeyard.base_command import CommandException
from kubeyard.commands.devel import BaseDevelCommand

logger = logging.getLogger(__name__)


class ShellCommand(BaseDevelCommand):
    """
    Command allows you to exec into container.
    """
    custom_script_name = 'shell'
    context_vars = ['pod', 'shell']

    def __init__(self, *, shell, pod, container, root, **kwargs):
        super().__init__(**kwargs)
        self.shell = shell
        self.pod = pod
        self.container = container
        self.root = root

    def run_default(self):
        try:
            sh.kubectl.exec(
                "-it",
                self.pod_name,
                "-c", self.container_name,
                '--',
                self.shell,
                "-c", self.before_command,
                _fg=True,
            )
        except sh.ErrorReturnCode_130:
            # Command exited using Ctrl+D or Ctrl+C
            pass
        finally:
            if self.after_command:
                sh.kubectl.exec(
                    self.pod_name,
                    "-c", self.container_name,
                    "--",
                    self.shell,
                    "-c", self.after_command,
                )

    @cached_property
    def pod_name(self) -> str:
        if self.pod:
            all_pods = sh.kubectl.get.pods('-o', 'jsonpath={.items[*].metadata.name}').split()
            # Exact match
            if self.pod in all_pods:
                return self.pod
            # Starting-with match
            pods = [pod for pod in all_pods if pod.startswith(self.pod)]
            pods.sort(key=len)
            if len(pods) == 0:
                raise CommandException(f"Not found pod equal or starting with '{self.pod}'")
            if len(pods) > 1:
                logger.warning(f"Found more than one pod. Using '{pods[0]}'")
            return pods[0]
        else:
            for pod in sh.kubectl.get.pods(_iter='out'):
                if self.image_name in pod:
                    return pod.split()[0]
        raise CommandException("Container not found, please specify container or fix project setup.")

    @cached_property
    def container_name(self) -> str:
        if self.container:
            return self.container
        else:
            return self.image_name

    @cached_property
    def username(self) -> str:
        return str(sh.whoami()).strip()

    @property
    def before_command(self):
        if self.root:
            return self.shell
        return (
            'groupadd -f -g {gid} {username}; '
            'adduser -q --gecos "" --disabled-password --uid {uid} --gid {gid} {username}; '
            'su {username}; '
        ).format(
            gid=self.gid,
            uid=self.uid,
            username=self.username,
        )

    @property
    def after_command(self) -> str:
        if self.root:
            return ""
        else:
            return "userdel --remove {username}; ".format(username=self.username)



import contextlib
import logging
import pathlib

import yaml

from kubeyard import ascii_art
from kubeyard import base_command
from kubeyard import context_factories
from kubeyard import dependencies
from kubeyard import kubernetes
from kubeyard import settings

logger = logging.getLogger(__name__)


class GlobalCommand(base_command.BaseCommand):
    @property
    def context(self):
        return context_factories.GlobalContextFactory().get()


class SetupCommand(GlobalCommand):
    """
    Set up global context. It should be run after instalation of kubeyard.
    If not provided, kubeyard will be guessing. If minikube is installed, development context will be used.
    If minikube will be not detected, kubeyard setup production context.
    Production context requires configured docker and kubectl. Also needs your microservices secrets to be at
    ~/kubernetes_secrets/.
    """

    def __init__(self, *, mode):
        self.mode = mode

    def run(self):
        user_context = self.get_current_user_context()
        if 'KUBEYARD_GLOBAL_SECRETS' not in user_context:
            user_context['KUBEYARD_GLOBAL_SECRETS'] = str(self.default_global_secrets_directory)
        if 'KUBEYARD_LOG_LEVEL' not in user_context:
            user_context['KUBEYARD_LOG_LEVEL'] = settings.DEFAULT_KUBEYARD_LOG_LEVEL
        if 'KUBEYARD_VM_DRIVER' not in user_context:
            user_context['KUBEYARD_VM_DRIVER'] = settings.DEFAULT_KUBEYARD_VM_DRIVER
        kubeyard_mode = self.get_kubeyard_mode()
        user_context['KUBEYARD_MODE'] = kubeyard_mode
        self.print_info(kubeyard_mode)
        with self.user_context_filepath.open('w') as context_file:
            yaml.dump(dict(user_context), stream=context_file, default_flow_style=False)
        new_context = dict(self.context, **user_context)
        kubernetes.setup_cluster_context(new_context)

    def get_current_user_context(self):
        return context_factories.GlobalContextFactory().user_context

    def get_kubeyard_mode(self):
        minikube_installed = dependencies.is_command_available('minikube')
        if self.mode == 'production':
            if minikube_installed:
                logger.error('Minikube installed! Use --development option')
                exit(1)
            else:
                return 'production'
        elif self.mode == 'development':
            if not minikube_installed:
                logger.error('Minikube not installed!')
                exit(1)
            else:
                return 'development'
        elif self.mode is None:
            return 'development' if minikube_installed else 'production'

    def print_info(self, kubeyard_mode):
        ascii_art.print_ascii_art()
        logger.info('Setting up {} mode...'.format(kubeyard_mode))

    @property
    def user_context_filepath(self):
        user_context_filepath = pathlib.Path(self.context['KUBEYARD_USER_CONTEXT_FILEPATH'])
        with contextlib.suppress(FileExistsError):
            user_context_filepath.parent.mkdir()
        return user_context_filepath

    @property
    def default_global_secrets_directory(self):
        global_secrets = self.user_context_filepath.parent / 'global-secrets/'
        with contextlib.suppress(FileExistsError):
            global_secrets.mkdir()
        return global_secrets


class InstallGlobalSecretsCommand(GlobalCommand):
    """
    Installs secrets from global secrets directory (usualy from ~/.kubeyard/global-secrets/).
    Usually secrets are maintained per microservice.
    Some secrets are easier to maintain if they are in one global file.

    In example: redis databases.
    """

    def run(self):
        kubernetes.install_global_secrets(self.context)



from pprint import pprint

from kubeyard import base_command


class DebugCommand(base_command.InitialisedRepositoryCommand):
    """
    Prints current context.

    If NAME supplied prints value of the single variable instead of dict containing whole context.
    """
    def __init__(self, name, **kwargs):
        super().__init__(**kwargs)
        self.name = name

    def run(self):
        super().run()
        if self.name is not None:
            print(self.context.get(self.name.upper()))
        else:
            pprint(self.context)



import os

DEFAULT_KUBEYARD_PROJECT_DIR = '.'
DEFAULT_KUBEYARD_SCRIPTS_DIR = 'scripts'
DEFAULT_SWCLI_CONTEXT_FILEPATH = 'config/sw_cli.yml'  # TODO: remove legacy
DEFAULT_KUBEYARD_CONTEXT_FILEPATH = 'config/kubeyard.yml'
DEFAULT_SWCLI_USER_CONTEXT_FILEPATH = '.sw_cli/context.yml'  # TODO: remove legacy
DEFAULT_KUBEYARD_USER_CONTEXT_FILEPATH = '.kubeyard/context.yml'
DEFAULT_KUBEYARD_LOG_LEVEL = 'INFO'
DEFAULT_KUBEYARD_VM_DRIVER = 'docker'
DEFAULT_KUBERNETES_DEPLOY_DIR = 'config/kubernetes/deploy'
DEFAULT_KUBERNETES_DEV_DEPLOY_OVERRIDES_DIR = 'config/kubernetes/development_overrides'
DEFAULT_KUBERNETES_DEV_SECRETS_DIR = 'config/kubernetes/dev_secrets'
KUBERNETES_PROD_SECRETS_DIR = (
        os.getenv('KUBEYARD_KUBERNETES_PROD_SECRETS_DIR') or
        os.getenv('SW_CLI_KUBERNETES_PROD_SECRETS_DIR') or  # TODO: remove legacy
        'kubernetes_secrets'
)
DEFAULT_DOCKER_REGISTRY_NAME = 'registry.hub.docker.com'
DEFAULT_PROJECT_NAME_PATTERN = '{project_name}'
DEFAULT_KUBE_SERVICE_NAME_PATTERN = '{dashed_project_name}'
DEFAULT_KUBE_SERVICE_PORT = '80'
DEFAULT_KUBE_LIVE_RELOAD_PORT = '30555'
DEFAULT_DEV_POSTGRES_NAME = 'kubernetes-postgres'
DEFAULT_DEV_PUBSUB_NAME = 'kubernetes-pubsub'
DEFAULT_DEV_REDIS_NAME = 'kubernetes-redis'
DEFAULT_DEV_ELASTIC_NAME = 'kubernetes-es'
DEFAULT_DEV_CASSANDRA_NAME = 'kubernetes-cassandra'
DEFAULT_DEV_TLD = 'testing'
DEFAULT_DEV_DOMAINS = ()
DEFAULT_TEST_DATABASE_IMAGE = 'postgres:10.3'
DEFAULT_TEST_DATABASE_NAME = 'test'
DEFAULT_TEST_MIGRATION_COMMAND = 'migrate_for_tests'
DEFAULT_TEST_COMMAND = 'run_tests'



import logging
import pathlib
import tempfile

import sh

import kubeyard.files_generator

from kubeyard import base_command

logger = logging.getLogger(__name__)


class InstallCompletion(base_command.BaseCommand):
    """
    Run this to enable bash completion. It writes /etc/bash_completion.d/kubeyard file.
    It will work automatically on your next bash use.

    You can enable it in your current sessions by running `. /etc/bash_completion.d/kubeyard`
    """
    completion_dst = pathlib.Path('/etc/bash_completion.d/kubeyard')

    def __init__(self, *, force):
        self.force = force

    def run(self):
        logger.info("Installing kubeyard completion...")
        if self.completion_dst.exists() and not self.force:
            logger.warning('File {} already exists. Skipping.'.format(str(self.completion_dst)))
        else:
            with tempfile.NamedTemporaryFile(mode='r') as f:
                completion_dst_tmp = pathlib.Path(f.name)
                kubeyard.files_generator.copy_template('kubeyard-completion.sh', completion_dst_tmp, replace=True)
                completion_dst_tmp.chmod(0o644)
                with sh.contrib.sudo(_with=True):
                    sh.cp(str(completion_dst_tmp), str(self.completion_dst))



import logging
import logging.config


def init_logging():
    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'simple': {
                'format': '%(asctime)s [%(levelname)s] %(name)s %(message)s',
                'datefmt': '%H:%M:%S',
            },
            'colored': {
                '()': 'colorlog.ColoredFormatter',
                'fmt': '%(asctime)s %(log_color)s[%(levelname)s]%(reset)s %(message)s',
                'datefmt': '%H:%M:%S',
                'log_colors': {
                    'DEBUG':    'cyan',
                    'INFO':     'white',
                    'WARNING':  'yellow',
                    'ERROR':    'red',
                    'CRITICAL': 'red,bold_red',
                },
            },
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'level': logging.DEBUG,
                'formatter': 'simple',
            },
            'console-colored': {
                'class': 'logging.StreamHandler',
                'level': logging.DEBUG,
                'formatter': 'colored',
            },
        },
        'root': {
            'level': logging.INFO,
            'handlers': ['console'],
        },
        'loggers': {
            'kubeyard': {
                'level': logging.INFO,
                'handlers': ['console-colored'],
                'propagate': False,
            },
            'sh.command': {
                'level': logging.WARNING,
            },
        },
    })



import logging
import pathlib

import sh

from kubeyard import dependencies
from kubeyard import kubernetes

logger = logging.getLogger(__name__)
definitions_directory = pathlib.Path(__file__).parent.parent / 'definitions' / 'dev_requirements'


class Requirement:
    valid_arguments = ()

    def __init__(self, context: dict):
        self.context = context

    def __call__(self, arguments: dict):
        if all(key in self.valid_arguments for key in arguments.keys()):
            self.run(arguments)
        else:
            logger.warning(
                'Requirement configuration is not valid: {}\n'
                'Available options are: {}'.format(arguments, self.valid_arguments))

    def run(self, arguments: dict):
        raise NotImplementedError


class Postgres(Requirement):
    valid_arguments = ('name')

    def run(self, arguments: dict):
        database_name = arguments.get('name') or self.context['KUBE_SERVICE_NAME']
        dependency = PostgresDependency()
        dependency.ensure_running()
        dependency.ensure_database_present(database_name)


class PostgresDependency(dependencies.KubernetesDependency):
    name = 'dev-postgres'
    definition = definitions_directory / 'postgres.yaml'
    started_log = 'PostgreSQL init process complete; ready for start up.'

    def ensure_database_present(self, database_name):
        logger.debug('Ensuring that database "{}" exists...'.format(database_name))
        try:
            self.run_command('createdb', database_name, '-U', 'postgres')
        except sh.ErrorReturnCode as e:
            if b'already exists' not in e.stderr:
                raise e
            else:
                logger.debug('Database "{}" exists'.format(database_name))
        else:
            logger.debug('Database "{}" created'.format(database_name))


class CockroachDB(Requirement):
    valid_arguments = ('name',)

    def run(self, arguments: dict):
        database_name = arguments.get('name') or self.context['KUBE_SERVICE_NAME']
        dependency = CockroachDBDependency()
        dependency.ensure_running()
        dependency.ensure_database_present(database_name)


class CockroachDBDependency(dependencies.KubernetesDependency):
    name = 'dev-cockroachdb'
    definition = definitions_directory / 'cockroachdb.yaml'
    started_log = 'CockroachDB node starting'

    def ensure_database_present(self, database_name):
        logger.debug('Ensuring that database "{}" exists...'.format(database_name))
        try:
            self.run_command('/cockroach/cockroach', 'sql', '--insecure',
                             '-e', 'CREATE DATABASE "{}";'.format(database_name))
        except sh.ErrorReturnCode as e:
            if b'already exists' not in e.stderr:
                raise e
            else:
                logger.debug('Database "{}" exists'.format(database_name))
        else:
            logger.debug('Database "{}" created'.format(database_name))


class Elasticsearch(Requirement):
    valid_arguments = ()

    def run(self, arguments: dict):
        self.ensure_elastic_running()

    def ensure_elastic_running(self):
        ElasticsearchDependency().ensure_running()


class ElasticsearchDependency(dependencies.KubernetesDependency):
    name = 'dev-elasticsearch'
    definition = definitions_directory / 'elasticsearch.yaml'
    started_log = '] started'


class PubSubEmulator(Requirement):
    valid_arguments = ('topic', 'subscription')

    def run(self, arguments: dict):
        topic_name = arguments.get('topic') or self.context['KUBE_SERVICE_NAME']
        dependency = PubSubDependency()
        dependency.ensure_running()
        dependency.ensure_topic_present(topic_name)
        try:
            subscription_name = arguments['subscription']
        except KeyError:
            logger.debug("Subscription not specified, it won't be created")
        else:
            dependency.ensure_subscription_present(topic_name, subscription_name)


class PubSubDependency(dependencies.KubernetesDependency):
    name = 'dev-pubsub'
    definition = definitions_directory / 'pubsub-emulator.yaml'
    started_log = '[pubsub] INFO: Server started, listening on'

    def ensure_topic_present(self, topic_name):
        logger.debug('Ensuring that topic "{}" exists...'.format(topic_name))
        try:
            self.run_command('pubsub_add_topic', topic_name)
        except sh.ErrorReturnCode as e:
            if b'Topic already exists' not in e.stderr:
                raise
            else:
                logger.debug('Topic "{}" exists'.format(topic_name))
        else:
            logger.debug('Topic "{}" created'.format(topic_name))

    def ensure_subscription_present(self, topic_name, subscription_name):
        logger.debug('Ensuring that subscription "{}" exists...'.format(subscription_name))
        try:
            self.run_command('pubsub_add_subscription', topic_name, subscription_name)
        except sh.ErrorReturnCode as e:
            if b'Subscription already exists' not in e.stderr:
                raise
            else:
                logger.debug('Subscription "{}" exists'.format(subscription_name))
        else:
            logger.debug('Subscription "{}" created'.format(subscription_name))


class Redis(Requirement):
    valid_arguments = ('name', )
    secret_name = 'redis-urls'

    def run(self, arguments: dict):
        dependency = RedisDependency()
        dependency.ensure_running()
        secret_key = arguments.get('name') or self.context['KUBE_SERVICE_NAME']
        secrets_manipulator = kubernetes.get_global_secrets_manipulator(self.context, self.secret_name)
        self.ensure_secret_is_present_in_file(secrets_manipulator, secret_key, redis_host=dependency.name)
        self.ensure_secret_is_installed(secrets_manipulator, secret_key)

    def ensure_secret_is_present_in_file(self, secrets_manipulator, secret_key, redis_host):
        logger.debug('Ensuring that secret key "{}" is present in file...'.format(secret_key))
        redis_urls = secrets_manipulator.get_literal_secrets_mapping()
        if secret_key in redis_urls:
            logger.debug('Secret key is already present in file')
        else:
            count = len(redis_urls)
            secrets_manipulator.set_literal_secret(
                key=secret_key, value='redis://{}:6379/{}'.format(redis_host, count))
            logger.debug('Secret key added to file')

    def ensure_secret_is_installed(self, secrets_manipulator, secret_key):
        logger.debug('Ensuring that secret key "{}" is present in secret "{}"...'.format(secret_key, self.secret_name))
        if secrets_manipulator.is_key_present(secret_key):
            logger.debug('Secret key is already present in secret')
        else:
            kubernetes.install_global_secrets(self.context)
            logger.debug('Secret key added to secret')


class RedisDependency(dependencies.KubernetesDependency):
    name = 'dev-redis'
    definition = definitions_directory / 'redis.yaml'
    started_log = 'The server is now ready to accept connections'


class Cassandra(Requirement):
    valid_arguments = ('keyspace')

    def run(self, arguments: dict):
        keyspace_name = arguments.get('keyspace') or self.context['KUBE_SERVICE_NAME']
        dependency = CassandraDependency()
        dependency.ensure_running()
        dependency.ensure_database_present(keyspace_name)


class CassandraDependency(dependencies.KubernetesDependency):
    name = 'dev-cassandra'
    definition = definitions_directory / 'cassandra.yaml'
    started_log = "Created default superuser role 'cassandra'"

    def ensure_database_present(self, keyspace_name):
        keyspace_name = self.clean_keyspace_name(keyspace_name)
        logger.debug('Ensuring that keyspace "{}" exists...'.format(keyspace_name))
        query = ("create keyspace %s with replication = {'class': 'SimpleStrategy', "
                 "'replication_factor': 1}" % keyspace_name)
        try:
            self.run_command('cqlsh', '-e', query)
        except sh.ErrorReturnCode as e:
            if b'already exists' not in e.stderr:
                raise e
            else:
                logger.debug('Keyspace "{}" exists'.format(keyspace_name))
        else:
            logger.debug('Keyspace "{}" created'.format(keyspace_name))

    def clean_keyspace_name(self, original):
        cleaned = original.replace('-', '_')
        if cleaned != original:
            logger.warning("Keyspace name can't contain dashes (-), so it's been changed to: %s" % cleaned)
        return cleaned


class RabbitMQ(Requirement):
    valid_arguments = ()

    def run(self, arguments: dict):
        dependency = RabbitMQDependency()
        dependency.ensure_running()


class RabbitMQDependency(dependencies.KubernetesDependency):
    name = 'dev-rabbitmq'
    definition = definitions_directory / 'rabbitmq.yaml'
    started_log = 'Starting RabbitMQ'


class RequirementsDispatcher:
    commands = {
        'postgres': Postgres,
        'cockroachdb': CockroachDB,
        'redis': Redis,
        'elastic': Elasticsearch,
        'pubsub': PubSubEmulator,
        'cassandra': Cassandra,
        'rabbitmq': RabbitMQ,
    }

    def __init__(self, context: dict):
        self.context = context

    def dispatch_all(self, requirements: dict):
        for requirement in requirements:
            if 'kind' in requirement:
                self.dispatch(requirement)
            else:
                logger.warning("Skipping requirement without specified kind. Requirement: {}".format(requirement))

    def dispatch(self, requirement: dict):
        arguments = requirement.copy()
        kind = arguments.pop('kind')
        logger.info('Checking requirement of kind "{}"...'.format(kind))
        try:
            command = self.commands[kind](self.context)
        except KeyError:
            logger.warning('Kind "{}" is not supported!'.format(kind))
        else:
            command(arguments)
            logger.info('Requirement of kind "{}" satisfied'.format(kind))



import collections
import contextlib
import logging
import pathlib
import socket

import sh
import yaml

from kubeyard import minikube
from kubeyard import settings

logger = logging.getLogger(__name__)


def setup_cluster_context(context):
    _get_kubernetes_commands(context).context_setup()


def install_secrets(context):
    logger.info('Installing secrets...')
    _get_kubernetes_commands(context).install_secrets()
    logger.info('Secrets installed')


def install_global_secrets(context):
    logger.info('Installing global secrets...')
    for path in pathlib.Path(context['KUBEYARD_GLOBAL_SECRETS']).iterdir():
        if path.is_dir():
            GlobalSecretsInstaller(context, path.name).install()
    logger.info('Global secrets installed')


def get_global_secrets_manipulator(context, secret_name):
    return KubernetesSecretsManipulator(
        secret_name,
        pathlib.Path(context['KUBEYARD_GLOBAL_SECRETS']) / secret_name,
    )


def _get_kubernetes_commands(context):
    if context['KUBEYARD_MODE'] == 'development':
        return KubernetesCommands(
            context_setup=DevelopmentKubernetesContext(context).setup,
            install_secrets=DevelopmentKubernetesSecretsInstaller(context).install,
        )
    else:
        return KubernetesCommands(
            context_setup=ProductionKubernetesContext().setup,
            install_secrets=ProductionKubernetesSecretsInstaller(context).install,
        )


KubernetesCommands = collections.namedtuple('KubernetesCommand', ['context_setup', 'install_secrets'])


class BaseKubernetesContext:
    def setup(self):
        with contextlib.suppress(sh.ErrorReturnCode):
            sh.kubectl('delete', 'configmap', 'global')
        sh.kubectl('create', 'configmap', 'global',
                   '--from-literal', 'monolith-host={}'.format(self.monolith_host),
                   '--from-literal', 'base-domain={}'.format(self.base_domain),
                   '--from-literal', 'alternative-domain={}'.format(self.alternative_domain),
                   '--from-literal', 'debug={}'.format(self.debug),
                   )

    @property
    def monolith_host(self):
        raise NotImplementedError

    @property
    def base_domain(self):
        raise NotImplementedError

    @property
    def alternative_domain(self):
        raise NotImplementedError

    @property
    def debug(self):
        raise NotImplementedError


class DevelopmentKubernetesContext(BaseKubernetesContext):
    base_domain = 'testing'
    alternative_domain = 'pl-testing'
    debug = 'True'

    def __init__(self, context):
        self.cluster = minikube.ClusterFactory().get(context)

    def setup(self):
        self.cluster.ensure_started()
        super().setup()

    @property
    def monolith_host(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        return s.getsockname()[0]


class ProductionKubernetesContext(BaseKubernetesContext):
    debug = 'False'
    base_domain = 'socialwifi.com'
    alternative_domain = 'socialwifi.pl'
    monolith_host = 'socialwifi.com'


class KubernetesSecretsManipulator:
    def __init__(self, secret_name, secrets_path):
        self.secret_name = secret_name
        self.secrets_path = secrets_path

    @property
    def yml_source_path(self):
        return self.secrets_path / 'secrets.yml'

    def get_literal_secrets(self):
        return self.get_literal_secrets_mapping().items()

    def get_file_secrets(self):
        self._ensure_path_exists()
        for subpath in self.secrets_path.iterdir():
            if subpath != self.yml_source_path:
                yield subpath

    def set_literal_secret(self, key, value):
        self._ensure_path_exists()
        literal_secrets = self.get_literal_secrets_mapping()
        literal_secrets[key] = value
        with self.yml_source_path.open('w+') as yml_source:
            yaml.dump(literal_secrets, yml_source)

    def get_literal_secrets_mapping(self):
        if self.yml_source_path.exists():
            with self.yml_source_path.open() as yml_source:
                secrets = yaml.safe_load(yml_source)
            return secrets or {}
        else:
            return {}

    def _ensure_path_exists(self):
        with contextlib.suppress(FileExistsError):
            self.secrets_path.mkdir(parents=True)

    def is_key_present(self, key):
        try:
            yml_output = str(sh.kubectl(
                'get', 'secrets', self.secret_name,
                '--output', 'yaml',
            ))
        except sh.ErrorReturnCode:
            return False
        else:
            secret = yaml.safe_load(yml_output)
            return key in secret['data']


class BaseKubernetesSecretsInstaller:
    def __init__(self, context):
        self.context = context

    def install(self):
        command = ['create', 'secret', 'generic', self.secret_name, '--dry-run', '-o', 'yaml']
        literal_secrets = list(self.manipulator.get_literal_secrets())
        file_secrets = list(self.manipulator.get_file_secrets())
        if literal_secrets or file_secrets:
            for key, value in literal_secrets:
                command.append('--from-literal={}={}'.format(key, value))
            for subpath in file_secrets:
                command.append('--from-file={}'.format(subpath))
            sh.kubectl(sh.kubectl(*command), 'apply', '--record', '-f', '-')

    @property
    def manipulator(self):
        return KubernetesSecretsManipulator(self.secret_name, self.secrets_path)

    @property
    def secret_name(self):
        raise NotImplementedError

    @property
    def secrets_path(self):
        raise NotImplementedError


class BaseProjectKubernetesSecretsInstaller(BaseKubernetesSecretsInstaller):
    @property
    def secret_name(self):
        return self.context['KUBE_SERVICE_NAME']


class DevelopmentKubernetesSecretsInstaller(BaseProjectKubernetesSecretsInstaller):
    @property
    def secrets_path(self):
        return pathlib.Path(self.context['PROJECT_DIR'])/self.context['KUBERNETES_DEV_SECRETS_DIR']


class ProductionKubernetesSecretsInstaller(BaseProjectKubernetesSecretsInstaller):
    @property
    def secrets_path(self):
        return pathlib.Path.home() / settings.KUBERNETES_PROD_SECRETS_DIR / self.secret_name


class GlobalSecretsInstaller(BaseKubernetesSecretsInstaller):
    secret_name = None

    def __init__(self, context, secret_name):
        super().__init__(context)
        self.secret_name = secret_name

    @property
    def secrets_path(self):
        return pathlib.Path(self.context['KUBEYARD_GLOBAL_SECRETS']) / self.secret_name



import collections
import logging
import pathlib
import re

import yaml

from cached_property import cached_property

from kubeyard import io_utils
from kubeyard import settings

logger = logging.getLogger(__name__)


class Context(dict):
    def as_environment(self):
        for key, value in self.items():
            if isinstance(value, str):
                yield key, value
            else:
                yield key, yaml.dump(value)


class GlobalContextFactory:
    def __init__(self):
        self.user_context_path = get_user_context_path()

    def get(self):
        context = Context()
        context.update(self.base_user_context)
        context.update(self.user_context)
        return upper_keys(context)

    @cached_property
    def base_user_context(self):
        return Context({
            'KUBEYARD_USER_CONTEXT_FILEPATH': str(self.user_context_path),
            'KUBEYARD_MODE': 'production',
            'DEV_POSTGRES_NAME': settings.DEFAULT_DEV_POSTGRES_NAME,
            'DEV_PUBSUB_NAME': settings.DEFAULT_DEV_PUBSUB_NAME,
            'DEFAULT_DEV_ELASTIC_NAME': settings.DEFAULT_DEV_ELASTIC_NAME,
            'DEV_REDIS_NAME': settings.DEFAULT_DEV_REDIS_NAME,
            'DEV_CASSANDRA_NAME': settings.DEFAULT_DEV_CASSANDRA_NAME,
        })

    @cached_property
    def user_context(self):
        if self.user_context_path.exists():
            return load_context(self.user_context_path)
        else:
            return Context()


class BaseRepoContextFactory:
    def __init__(self, project_dir: str):
        self.project_dir = project_dir
        self.user_context_path = get_user_context_path()

    def get(self):
        context = Context({
            'PROJECT_NAME': self.default_project_name,
        })
        context.update(self.project_context)
        context.update(GlobalContextFactory().get())
        context['UNDERSCORED_PROJECT_NAME'] = self.normalize_name(context['PROJECT_NAME'])
        context['DASHED_PROJECT_NAME'] = self.normalize_name(context['PROJECT_NAME'], '-')
        return upper_keys(context)

    @property
    def project_context(self):
        raise NotImplementedError

    @cached_property
    def default_project_name(self) -> str:
        return pathlib.Path(self.project_dir).resolve().name

    @staticmethod
    def normalize_name(name: str, char='_') -> str:
        return re.sub(r'[-_\s]', char, name)


def get_user_context_path():
    """
    Helper for legacy sw-cli support.
    TODO: remove legacy
    """
    user_context_path = pathlib.Path.home() / settings.DEFAULT_KUBEYARD_USER_CONTEXT_FILEPATH
    if not user_context_path.exists():
        legacy_context_path = pathlib.Path.home() / settings.DEFAULT_SWCLI_USER_CONTEXT_FILEPATH
        if legacy_context_path.exists():
            logger.warning("Using legacy user config file: {}. Please rename it to: {}!".format(
                settings.DEFAULT_SWCLI_USER_CONTEXT_FILEPATH,
                settings.DEFAULT_KUBEYARD_USER_CONTEXT_FILEPATH,
            ))
            user_context_path = legacy_context_path
    return user_context_path


PromptedContext = collections.namedtuple('PromptedContext', ['variable', 'prompt', 'default'])


class EmptyRepoContextFactory(BaseRepoContextFactory):
    def __init__(self, project_dir, prompted_context):
        super().__init__(project_dir)
        self.prompted_context = prompted_context

    @cached_property
    def project_context(self):
        context = {}
        for prompt_config in self.prompted_context:
            context[prompt_config.variable] = io_utils.default_input(
                prompt_config.prompt,
                prompt_config.default.format(
                    project_name=context.get('PROJECT_NAME', self.default_project_name),
                    underscored_project_name=self.normalize_name(
                        context.get('PROJECT_NAME', self.default_project_name), '_',
                    ),
                    dashed_project_name=self.normalize_name(
                        context.get('PROJECT_NAME', self.default_project_name), '-',
                    ),
                ),
            )
        context['DOCKER_IMAGE_NAME'] = self.normalize_name(context.get('PROJECT_NAME', self.default_project_name))
        context['DOCKER_REGISTRY_DOMAIN'] = context['DOCKER_REGISTRY_NAME'].split('/', 1)[0]
        return context


class InitialisedRepoContextFactory(BaseRepoContextFactory):
    def __init__(self, project_dir, context_filepath=settings.DEFAULT_KUBEYARD_CONTEXT_FILEPATH):
        super().__init__(project_dir)
        self.context_filepath = context_filepath
        self.filename = project_dir / context_filepath
        if not self.filename.exists() and context_filepath == settings.DEFAULT_KUBEYARD_CONTEXT_FILEPATH:
            legacy_filename = project_dir / settings.DEFAULT_SWCLI_CONTEXT_FILEPATH
            if legacy_filename.exists():
                # TODO: remove legacy
                logger.warning("Using legacy project config file: {} file. Please rename it to: {}!".format(
                    settings.DEFAULT_SWCLI_CONTEXT_FILEPATH,
                    settings.DEFAULT_KUBEYARD_CONTEXT_FILEPATH,
                ))
                self.filename = legacy_filename
        if not self.filename.exists():
            raise FileNotFoundError("File does not exist: %s" % self.filename)

    @cached_property
    def project_context(self):
        context = Context({
            'PROJECT_DIR': str(self.project_dir),
            'KUBEYARD_CONTEXT_FILEPATH': self.context_filepath,
            'KUBEYARD_SCRIPTS_DIR': settings.DEFAULT_KUBEYARD_SCRIPTS_DIR,
            'KUBERNETES_DEV_SECRETS_DIR': settings.DEFAULT_KUBERNETES_DEV_SECRETS_DIR,
            'DEV_TLD': settings.DEFAULT_DEV_TLD,
            'DEV_DOMAINS': settings.DEFAULT_DEV_DOMAINS,
            'TEST_DATABASE_IMAGE': settings.DEFAULT_TEST_DATABASE_IMAGE,
            'TEST_DATABASE_NAME': settings.DEFAULT_TEST_DATABASE_NAME,
            'TEST_MIGRATION_COMMAND': settings.DEFAULT_TEST_MIGRATION_COMMAND,
            'TEST_COMMAND': settings.DEFAULT_TEST_COMMAND,
        })
        context.update(self.saved_context)
        return context

    @cached_property
    def saved_context(self):
        return load_context(self.filename)


def load_context(path):
    with path.open() as fp:
        return de_legacy(upper_keys(yaml.safe_load(fp)))


def upper_keys(d):
    ret = Context()
    for k, v in d.items():
        ret.update({k.upper(): v})
    return ret


def de_legacy(context: Context) -> Context:
    """
    :param context: ready context with capital letters keys.
    :return: context with keys renamed from ..._SWCLI_.. to ..._KUBEYARD_...
    TODO: remove legacy
    """
    new_context = Context()
    for key, value in context.items():
        if 'SWCLI' in key:
            legacy_key = key
            key = key.replace('SWCLI', 'KUBEYARD')
            logger.warning("You have legacy entry: {} in context. Please rename it to: {}!".format(legacy_key, key))
        new_context.update({key: value})
    return new_context



import collections
import contextlib
import logging
import os
import pathlib
import shutil

import jinja2

logger = logging.getLogger(__name__)

templates_directory = pathlib.Path(__file__).parent / 'templates'
jinja_loader = jinja2.FileSystemLoader('/')
jinja_environment = jinja2.Environment(loader=jinja_loader)


def copy_template(template, destination, context=None, *, replace=False):
    context = context or {}
    source = templates_directory / template
    environment = CopyEnvironment(source, destination, context)
    for path in traverse(source):
        if path.is_dir():
            DirectoryTemplate(path, environment).render()
        else:
            if path.match('*.template'):
                file_template = TemplateFileTemplate(path, environment)
            else:
                file_template = SimpleFileTemplate(path, environment)
            if file_template.destination.exists() and not replace:
                logger.warning('File {} already exists. Skipping.'.format(str(file_template.destination)))
            else:
                file_template.render()


def recursive_chown(directory, uid):
    for path in traverse(directory):
        os.chown(str(path), uid=uid, gid=-1)


def traverse(path: pathlib.Path):
    yield path
    if path.is_dir():
        for sub in path.iterdir():
            yield from traverse(sub)


CopyEnvironment = collections.namedtuple('CopyEnvironment', ['source_root', 'destination_root', 'context'])


class BaseTemplate:
    def __init__(self, source, environment):
        self.source = source
        self.environment = environment

    @property
    def destination(self) -> pathlib.Path:
        relative_path = self.source.relative_to(self.environment.source_root)
        destination_template = jinja_environment.from_string(str(self.environment.destination_root / relative_path))
        destination = destination_template.render(self.environment.context)
        return pathlib.Path(destination)

    def render(self):
        raise NotImplementedError


class DirectoryTemplate(BaseTemplate):
    def render(self):
        with contextlib.suppress(FileExistsError):
            self.destination.mkdir()


class SimpleFileTemplate(BaseTemplate):
    def render(self):
        shutil.copy(str(self.source), str(self.destination))


class TemplateFileTemplate(BaseTemplate):
    def render(self):
        template = jinja_environment.get_template(str(self.source))
        rendered = template.render(self.environment.context)
        with self.destination.open('w') as destination_file:
            destination_file.write(rendered)

    @property
    def destination(self):
        return super().destination.with_suffix('')



from kubeyard.commands.devel import BaseDevelCommand


class PushCommand(BaseDevelCommand):
    """
    Runs `docker push` on docker image built by build command. It also tags image as latest adn push it as well.
    Can be overridden in <project_dir>/sripts/push.

    If kubeyard is set up in development mode it uses minikube as docker host.

    Normally you want to run it only in production.
    """
    custom_script_name = 'push'

    def run_default(self):
        self.docker_with_output('push', self.image)
        self.docker_with_output('tag', self.image, self.latest_image)
        self.docker_with_output('push', self.latest_image)



from kubeyard.commands.bash_completion import InstallCompletion
from kubeyard.commands.build import BuildCommand
from kubeyard.commands.custom_script import CustomScriptCommand
from kubeyard.commands.debug import DebugCommand
from kubeyard.commands.deploy import DeployCommand
from kubeyard.commands.fix_code_style import FixCodeStyleCommand
from kubeyard.commands.global_commands import InstallGlobalSecretsCommand
from kubeyard.commands.global_commands import SetupCommand
from kubeyard.commands.init import InitCommand
from kubeyard.commands.push import PushCommand
from kubeyard.commands.shell import ShellCommand
from kubeyard.commands.test import TestCommand
from kubeyard.commands.update_requirements import UpdateRequirementsCommand

__all__ = [
    InstallCompletion,
    CustomScriptCommand,
    DebugCommand,
    BuildCommand,
    DeployCommand,
    PushCommand,
    UpdateRequirementsCommand,
    FixCodeStyleCommand,
    InstallGlobalSecretsCommand,
    SetupCommand,
    InitCommand,
    TestCommand,
    ShellCommand,
]



import datetime
import getpass
import logging
import os
import pathlib
import re
import sys
import time

import sh

from cached_property import cached_property

from kubeyard import settings

logger = logging.getLogger(__name__)


class Cluster:
    docker_env_keys = ['DOCKER_TLS_VERIFY', 'DOCKER_HOST', 'DOCKER_CERT_PATH', 'MINIKUBE_ACTIVE_DOCKERD']
    minimum_minikube_version = (1, 32, 0)

    def ensure_started(self):
        if not self.is_running():
            self.start()

    def is_running(self):
        raise NotImplementedError

    def start(self):
        self._before_start()
        self._start()
        self._after_start()

    def _before_start(self):
        self._check_version()

    def _start(self):
        raise NotImplementedError

    def _after_start(self):
        pass

    def docker_env(self):
        return {}

    def get_mounted_project_dir(self, project_dir):
        raise NotImplementedError

    def _check_version(self):
        version = str(sh.minikube('version'))
        version_pattern = r'v(\d+)\.(\d+)\.(\d+)'
        match = re.search(version_pattern, version)
        if not match:
            logger.warning('Could not determine minikube version. Got: "{}"'.format(version))
        else:
            version = tuple(map(int, match.groups()))
            if version < self.minimum_minikube_version:
                logger.warning('Minikube version {} is not supported, you may run into issues. '
                               'Minimum supported version: {}'.format(version, self.minimum_minikube_version))


class DockerCluster(Cluster):
    kubernetes_version = 'v1.28'
    static_ip = '192.168.200.200'
    cpu_limit = 'no-limit'
    memory_limit = 'no-limit'

    def is_running(self):
        try:
            output = sh.minikube('status')
        except sh.ErrorReturnCode:
            return False
        else:
            return output.exit_code == 0

    def _start(self):
        logger.info('Starting minikube in Docker...')
        sh.minikube(
            'start',
            '--driver', 'docker',
            '--kubernetes-version', self.kubernetes_version,
            '--extra-config', 'apiserver.service-node-port-range=1-32767',
            '--cpus', self.cpu_limit,
            '--memory', self.memory_limit,
            '--mount',
            '--mount-string', '{}:{}'.format(os.environ['HOME'], os.environ['HOME']),
            '--static-ip', self.static_ip,
            _out=sys.stdout.buffer, _err=sys.stdout.buffer)

    def docker_env(self):
        variables = map(lambda x: '$' + x, self.docker_env_keys)
        result = sh.bash('-c', 'eval $(minikube docker-env); echo "%s"' % '|'.join(variables))
        values = result.strip("\n").split('|')
        return dict(zip(self.docker_env_keys, values))

    def get_mounted_project_dir(self, project_dir):
        return project_dir


class NativeLocalkubeCluster(Cluster):
    _docker_config_path = '/etc/docker/daemon.json'

    def __init__(self):
        self._docker_config_backup_path = f'{self._docker_config_path}_backup_{datetime.datetime.now().isoformat()}'

    def is_running(self):
        try:
            status = sh.systemctl('is-active', 'kubelet')
        except sh.ErrorReturnCode_3:
            return False
        else:
            return status.strip().lower() == 'active'

    def _before_start(self):
        super()._before_start()
        self._backup_docker_config()

    # minikube overrides contents of /etc/docker/daemon.json with settings that
    # shouldn't make any difference for us, but it breaks all custom modifications like "data-root"
    def _backup_docker_config(self):
        logger.info('Creating docker config file backup...')
        with sh.contrib.sudo(password=self._sudo_password, _with=True):
            sh.cp('-a', self._docker_config_path, self._docker_config_backup_path)
        logger.info('Docker config file backup created.')

    def _start(self):
        logger.info('Starting minikube without a VM...')
        sh.sudo('-E', '-S',
                *self._start_env_as_arguments,
                'minikube', 'start',
                '--driver', 'none',
                '--container-runtime', 'docker',
                '--kubernetes-version', 'v1.21.14',
                '--extra-config', 'apiserver.service-node-port-range=1-32767',
                _in=self._sudo_password, _out=sys.stdout.buffer, _err=sys.stdout.buffer)

    def _after_start(self):
        super()._after_start()
        self._restore_docker_config()

    def _restore_docker_config(self):
        logger.info('Restoring docker config file...')
        with sh.contrib.sudo(password=self._sudo_password, _with=True):
            sh.mv(self._docker_config_backup_path, self._docker_config_path)
            sh.systemctl('restart', 'docker.service')
        logger.info('(sleep 20s.) Docker config file restored, waiting for minikube to reconcile '
                    'after Docker restart...')
        time.sleep(20)
        logger.info('Waiting done.')

    def get_mounted_project_dir(self, project_dir):
        return project_dir

    @property
    def _start_env_as_arguments(self):
        env = {
            'CHANGE_MINIKUBE_NONE_USER': 'true',
        }
        return ['{}={}'.format(key, value) for key, value in env.items()]

    @cached_property
    def _sudo_password(self):
        prompt = "[sudo] password for %s: " % getpass.getuser()
        return getpass.getpass(prompt=prompt) + "\n"


class VirtualboxCluster(Cluster):
    def is_running(self):
        running_machines = sh.VBoxManage('list', 'runningvms')
        return 'minikube' in running_machines

    def _start(self):
        logger.info('Starting minikube with VirtualBox...')
        minikube_iso = 'https://storage.googleapis.com/minikube/iso/minikube-v0.23.4.iso'
        sh.minikube('start',
                    '--memory', '4096',
                    '--disk-size', '30g',
                    '--iso-url', minikube_iso,
                    '--docker-opt', 'storage-driver=overlay2',
                    _out=sys.stdout.buffer, _err=sys.stdout.buffer)

    def _after_start(self):
        super()._after_start()
        self._increase_inotify_limit()
        self._ensure_hosthome_mounted()

    def _increase_inotify_limit(self):
        sh.minikube('ssh', 'sudo sysctl fs.inotify.max_user_watches=16382')

    def _ensure_hosthome_mounted(self):
        if '/hosthome' not in sh.minikube('ssh', 'mount'):
            logger.info("Preparing hosthome directory...")
            try:
                sh.minikube('ssh', 'sudo mkdir /hosthome')
            except sh.ErrorReturnCode_1 as e:
                if b'can\'t create directory \'/hosthome\': File exists' not in e.stderr:
                    raise
            sh.minikube('ssh', 'sudo chmod 777 /hosthome')
            sh.minikube('ssh', 'sudo mount -t vboxsf -o uid=$(id -u),gid=$(id -g) hosthome /hosthome')

    def get_mounted_project_dir(self, project_dir):
        return pathlib.Path('/hosthome') / project_dir.relative_to('/home')

    def docker_env(self):
        variables = map(lambda x: '$' + x, self.docker_env_keys)
        result = sh.bash('-c', 'eval $(minikube docker-env); echo "%s"' % '|'.join(variables))
        values = result.strip("\n").split('|')
        return dict(zip(self.docker_env_keys, values))


class ClusterFactory:
    VM_DRIVERS = {
        'none': NativeLocalkubeCluster,
        'docker': DockerCluster,
        'virtualbox': VirtualboxCluster,
    }

    def get(self, context):
        vm_driver = context.get('KUBEYARD_VM_DRIVER', settings.DEFAULT_KUBEYARD_VM_DRIVER)
        return self.VM_DRIVERS[vm_driver]()



from pyfiglet import figlet_format
from termcolor import cprint

TEXT = 'KubeYard'
FONT_COLOR = 'cyan'
FONT = 'standard'


def print_ascii_art():
    ascii_art = figlet_format(TEXT, font=FONT, justify='center')
    cprint(ascii_art, FONT_COLOR, attrs=['bold'])



import abc
import logging
import os
import re
import signal
import sys
import typing

from contextlib import contextmanager

import sh

from cached_property import cached_property

from kubeyard import base_command
from kubeyard import minikube
from kubeyard import settings
from kubeyard.commands import custom_script

logger = logging.getLogger(__name__)

MAX_JOB_RETRIES = 2


class BaseDevelCommand(base_command.InitialisedRepositoryCommand):
    context_vars = ['image_name', 'tag']

    def __init__(self, *, use_default_implementation, image_name, tag, **kwargs):
        super().__init__(**kwargs)
        self._image_name = image_name
        self._tag = tag
        self.use_default_implementation = use_default_implementation
        if self.is_development:
            self.cluster = self._prepare_cluster(self.context)
            self.context.update(self.cluster.docker_env())
            self.context['HOST_VOLUMES'] = ' '.join(self.volumes)
        self.docker_runner = DockerRunner(self.context)

    @staticmethod
    def _prepare_cluster(context):
        logger.info('Checking if cluster is running and configured...')
        cluster = minikube.ClusterFactory().get(context)
        cluster.ensure_started()
        logger.info('Cluster is ready')
        return cluster

    @property
    def args(self) -> list:
        return []

    def run(self):
        super().run()
        custom_script_runner = custom_script.CustomScriptRunner(self.project_dir, self.custom_command_context)
        custom_script_exists = custom_script_runner.exists(self.custom_script_name)
        if self.use_default_implementation or not custom_script_exists:
            self.run_default()
        else:
            custom_script_runner.run(self.custom_script_name, self.args)

    def docker(self, *args, **kwargs):
        return self.docker_runner.run(*args, **kwargs)

    def docker_with_output(self, *args, **kwargs):
        return self.docker_runner.run_with_output(*args, **kwargs)

    @property
    def image(self):
        return '{}/{}:{}'.format(self.docker_repository, self.image_name, self.tag)

    @property
    def latest_image(self):
        return '{}/{}:latest'.format(self.docker_repository, self.image_name)

    @property
    def docker_repository(self):
        return self.context.get("DOCKER_REGISTRY_NAME") or settings.DEFAULT_DOCKER_REGISTRY_NAME

    @property
    def image_name(self):
        return self._image_name or self.context["DOCKER_IMAGE_NAME"]

    @property
    def tag(self):
        return self._tag or self.default_tag

    @property
    def default_tag(self):
        if self.is_development:
            return 'dev'
        else:
            return 'latest'

    @property
    def is_development(self):
        return self.context['KUBEYARD_MODE'] == 'development'

    def run_default(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def custom_script_name(self):
        raise NotImplementedError

    @property
    def volumes(self) -> typing.Iterable[str]:
        if self.is_development:
            mounted_project_dir = self.cluster.get_mounted_project_dir(self.project_dir)
            for volume in self.context.get('DEV_MOUNTED_PATHS', []):
                if 'mount-in-tests' in volume and volume['mount-in-tests']['image-name'] == self.image_name:
                    host_path = str(mounted_project_dir / volume['host-path'])
                    container_path = volume['mount-in-tests']['path']
                    mount_mode = self.get_mount_mode(volume['mount-in-tests'])
                    yield from ['-v', '{}:{}:{}'.format(host_path, container_path, mount_mode)]

    def get_mount_mode(self, configuration):
        mount_mode = configuration.get('mount-mode', 'ro')
        if mount_mode not in {'ro', 'rw'}:
            raise base_command.CommandException('Volume "mount-mode" should be one of: "ro", "rw".')
        return mount_mode

    @cached_property
    def _id(self):
        return str(sh.id())

    @cached_property
    def uid(self) -> str:
        return re.findall(r"uid=(\d+)", self._id)[0]

    @cached_property
    def gid(self) -> str:
        return re.findall(r"gid=(\d+)", self._id)[0]


class DockerRunner:
    def __init__(self, context):
        self.context = context

    def run(self, *args, **kwargs):
        if self.run_can_be_waited(*args, **kwargs):
            process: sh.RunningCommand = sh.docker(*args, _env=self.sh_env, _bg_exc=False, **kwargs)
            try:
                process.wait()
            except KeyboardInterrupt as e:
                logger.info("Stopping running command...")
                process.signal(signal.SIGTERM)
                try:
                    children_pids = sh.ps('-o', 'pid', '--ppid', process.pid, '--no-headers').strip().split(' ')
                except sh.ErrorReturnCode_1:
                    pass
                else:
                    for pid in children_pids:
                        os.kill(int(pid), signal.SIGTERM)
                raise e
        else:
            process: sh.RunningCommand = sh.docker(*args, _env=self.sh_env, **kwargs)
        return process

    def run_can_be_waited(self, *args, _piped=False, _iter=False, _iter_noblock=False, **kwargs) -> bool:
        """Check special cases when sh require to not use .wait() method"""
        return not any((_piped, _iter, _iter_noblock))

    def run_with_output(self, *args, **kwargs):
        return self.run(*args, _out=sys.stdout.buffer, _err=sys.stdout.buffer, **kwargs)

    @cached_property
    def sh_env(self):
        env = os.environ.copy()
        env.update(self.context.as_environment())
        return env

    @contextmanager
    def temporary_volume(self):
        volume_name = self.run('volume', 'create').strip()
        logger.debug('volume_name: {}'.format(volume_name))
        yield volume_name
        self.run('volume', 'remove', volume_name)



import logging
import os

from kubeyard import base_command

logger = logging.getLogger(__name__)


class CustomScriptCommand(base_command.InitialisedRepositoryCommand):
    """
    * Custom script defined in ./scripts/ directory. Running it with kubeyard add current context to environment.
    """

    def __init__(self, script_name, custom_script_args, **kwargs):
        super().__init__(**kwargs)
        self.script_name = script_name
        self.custom_script_args = list(custom_script_args)

    def run(self):
        super().run()
        logger.info('Running custom script command "{}"...'.format(self.script_name))
        CustomScriptRunner(self.project_dir, self.context).run(self.script_name, self.custom_script_args)


class CustomScriptRunner:
    def __init__(self, project_dir, context):
        self.project_dir = project_dir
        self.context = context

    def run(self, script_name, args):
        filepath = self.scripts_dir_path / script_name
        env = os.environ.copy()
        env.update(self.context.as_environment())
        os.execvpe(file=str(filepath), args=[str(filepath)] + args, env=env)

    def exists(self, script_name):
        return (self.scripts_dir_path / script_name).exists()

    @property
    def scripts_dir_path(self):
        return self.project_dir / self.context.get('KUBEYARD_SCRIPTS_DIR')



import logging
import shlex

from kubeyard.commands.devel import BaseDevelCommand

logger = logging.getLogger(__name__)


class BuildCommand(BaseDevelCommand):
    """
    Builds docker image required to run tests and deployment. Can be overridden in <project_dir>/sripts/build.
    If kubeyard is set up in development mode it uses minikube as docker host.
    """
    custom_script_name = 'build'
    context_vars = ['image_context', 'docker_args']

    def __init__(self, *, image_context, docker_args, **kwargs):
        super().__init__(**kwargs)
        self.image_context = image_context
        self.docker_args = docker_args or ''

    def run_default(self):
        image_context = self.image_context or "{0}/docker".format(self.project_dir)
        logger.info('Building image "{}"...'.format(self.image))
        self.docker_with_output('build', '-t', self.image, *shlex.split(self.docker_args),  image_context)



import os
import pathlib

import click

from cached_property import cached_property

from kubeyard import settings
from kubeyard.commands import CustomScriptCommand


class CustomCommandsLoader(click.MultiCommand):
    scripts_dir = pathlib.Path(settings.DEFAULT_KUBEYARD_SCRIPTS_DIR)

    def __init__(self, main_cli: list, **attrs):
        """
        :param main_cli: Is required to pass main CLI to not override methods from CLI on current level.
        """
        super().__init__(**attrs)
        self.main_cli = main_cli

    @cached_property
    def custom_scripts(self):
        cmds = {}
        if self.scripts_dir.exists():
            scripts_dir = self.scripts_dir.resolve()
            for filepath in scripts_dir.glob("*"):
                if filepath.is_file() and self.is_executable(filepath):
                    cmds[filepath.name] = filepath
        return cmds

    @staticmethod
    def is_executable(filepath: pathlib.Path) -> bool:
        return os.access(filepath, os.X_OK)

    def list_commands(self, ctx):
        """Do not override commands from main CLI"""
        return [cs for cs in self.custom_scripts.keys() if cs not in self.main_cli.list_commands(ctx)]

    def get_command(self, ctx, cmd_name):
        from kubeyard.entrypoints.kubeyard import apply_common_options
        from kubeyard.entrypoints.kubeyard import initialized_repository_options

        @click.command(
            name=cmd_name,
            help=CustomScriptCommand.__doc__,
            context_settings=dict(
                ignore_unknown_options=True,
            ),
        )
        @apply_common_options(initialized_repository_options)
        @click.argument("custom_script_args", nargs=-1, type=click.UNPROCESSED)
        def _custom_script_command(**kwargs):
            CustomScriptCommand(script_name=self.custom_scripts[cmd_name], **kwargs).run()

        return _custom_script_command



import logging
import time

import sh

logger = logging.getLogger(__name__)


def is_command_available(name):
    try:
        sh.bash('which', name)
    except sh.ErrorReturnCode:
        return False
    else:
        return True


class KubernetesDependency:
    def ensure_running(self):
        logger.debug('Checking if container "{}" is running...'.format(self.name))
        if self.is_container_running():
            logger.debug('"{}" is running'.format(self.name))
        else:
            logger.debug('Starting "{}"...'.format(self.name))
            self.run_container()
            logger.debug('"{}" started'.format(self.name))

    def run_container(self):
        self._apply_definition()
        self._wait_until_ready()
        self._wait_for_started_log()

    def _apply_definition(self):
        sh.kubectl('apply', '--record', '-f', self.definition)
        try:
            sh.kubectl('expose', '-f', self.definition)
        except sh.ErrorReturnCode_1 as e:
            if b'already exists' not in e.stderr:
                raise e
            else:
                logger.debug('Service for "{}" exists'.format(self.name))

    def _wait_until_ready(self):
        logger.debug('Waiting for "{}" to start (possibly downloading image)...'.format(self.name))
        ready = False
        while not ready:
            ready = self.is_container_running()
            if not ready:
                time.sleep(1)
        logger.debug('"{}" started'.format(self.name))

    def _wait_for_started_log(self):
        logger.debug('Waiting for started log for "{}"...'.format(self.name))
        for log in sh.kubectl('logs', '-f', self.pod_name, _iter='out'):
            if self.started_log in log:
                break
        logger.debug('Started log for "{}" found'.format(self.name))

    def is_container_running(self):
        try:
            container_ready = str(sh.kubectl(
                'get', 'pods',
                '--selector', self.selector,
                '--output', 'jsonpath="{.items[*].status.containerStatuses[*].ready}"',
            )).strip()
        except sh.ErrorReturnCode as e:
            logger.debug(e)
            return False
        else:
            return container_ready == '"true"'

    def run_command(self, *args):
        return sh.kubectl('exec', self.pod_name, '--', *args)

    @property
    def pod_name(self):
        return str(sh.kubectl(
            'get', 'pods',
            '--output', 'custom-columns=NAME:.metadata.name',
            '--no-headers',
            '--selector', self.selector,
        )).strip()

    @property
    def selector(self):
        return 'app={}'.format(self.name)

    @property
    def started_log(self):
        raise NotImplementedError

    @property
    def name(self):
        raise NotImplementedError

    @property
    def definition(self):
        raise NotImplementedError



import abc
import json
import logging
import sys
import typing

import sh

from kubeyard.commands.devel import BaseDevelCommand
from kubeyard.commands.devel import DockerRunner

logger = logging.getLogger(__name__)


class TestCommand(BaseDevelCommand):
    """
    Runs tests in docker image built by build command. Can be overridden in <project_dir>/sripts/test.

    Supported databases:
        - Postgres
        - CockroachDB
    You can override migration command and test command passed to docker container with code.

    \b
    Example:
    test_migration_command: migrate_user
    test_command: db_tests_for_user

    If kubeyard is set up in development mode it uses minikube as docker host and mounts volumes configured in
    dev_mounted_paths in config/kubeyard.yml if they have mount-in-test set.

    \b
    Example:
    dev_mounted_paths:
    - name: dev-volume
      host-path: docker/source
      mount-in-tests:
        path: /package
        image-name: sw-project


    You can setup database before tests
    (on dev environment db will be cached, otherwise it will be removed immediately after tests).

    \b
    Example:
    tests_with_database: true
    test_database_type: postgres <- Available types: `postgres`, `cockroach`. `postgres` is default.
    test_database_image: postgres:10.3
    test_database_name: test

    """
    custom_script_name = 'test'
    context_vars = ["force_recreate_database", "force_migrate_database"]

    def __init__(self, *, force_recreate_database, force_migrate_database, test_options, **kwargs):
        super().__init__(**kwargs)
        self.test_options = test_options
        self.force_recreate_database = force_recreate_database
        self.force_migrate_database = force_migrate_database

    @property
    def args(self) -> list:
        return list(self.test_options)

    def run_default(self):
        if self.context.get('TESTS_WITH_DATABASE'):
            database_class = DATABASE_MAP[self.context.get("TEST_DATABASE_TYPE", "postgres")]
            with database_class(
                    is_development=self.is_development,
                    volumes=self.volumes,
                    context=self.context,
                    tag=self.tag,
                    docker_runner=self.docker_runner,
                    tested_image_name=self.image,
                    force_recreate=self.force_recreate_database,
                    force_migrate=self.force_migrate_database,
            ) as database:
                self.run_tests(database)
        else:
            self.run_tests()

    def run_tests(self, database: 'Database' = None):
        logger.info('Running tests...')
        try:
            self.docker_runner.run_with_output(
                'run',
                '--rm',
                '--init',
                '--net={}'.format(database.network if database else 'none'),
                *self.volumes,
                self.image,
                self.context['TEST_COMMAND'],
                *self.test_options,
            )
        except sh.ErrorReturnCode_1 as e:
            logger.debug(e)
            sys.exit(1)


class Database(metaclass=abc.ABCMeta):
    def __init__(
            self,
            is_development: bool,
            volumes: typing.Iterable[str],
            context: dict,
            tag: str,
            docker_runner: DockerRunner,
            tested_image_name: str,
            force_recreate: bool = False,
            force_migrate: bool = False,
    ):
        self.is_development = is_development
        self.volumes = volumes
        self.context = context
        self.tag = tag
        self.docker_runner = docker_runner
        self.tested_image_name = tested_image_name
        self.force_migrate = force_migrate
        self.force_recreate = force_recreate
        self._migrated = False

    def __enter__(self):
        if not self.is_development or self.force_recreate:
            self.remove_database()
        if self.container_stopped:
            logger.info("Found stopped DB, restarting it!")
            self.docker_runner.run('start', self.container_name)
        if not self.already_up:
            self.create()
            self.wait_until_ready()
            self.migrate()
        if self.force_migrate:
            self.migrate()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.is_development:
            self.remove_database()

    @property
    def container_stopped(self) -> bool:
        try:
            container_info = json.loads(str(self.docker_runner.run('inspect', self.container_name)))
        except sh.ErrorReturnCode_1:
            return False
        else:
            return container_info[0]['State']['Running'] is False

    @property
    def already_up(self) -> bool:
        try:
            self.docker_runner.run('inspect', self.container_name)
            return True
        except sh.ErrorReturnCode_1:
            return False

    @property
    def container_name(self) -> str:
        image_name = self.context['DOCKER_IMAGE_NAME']
        return f'db-test-{image_name}-{self.tag}'

    def remove_database(self):
        logger.info('Removing database...')
        try:
            self.docker_runner.run('rm', '--force', '--volumes', self.container_name)
        except sh.ErrorReturnCode_1 as e:
            logger.info('Database does not exist yet.')
            logger.debug(e)

    @abc.abstractmethod
    def create(self):
        raise NotImplementedError

    def wait_until_ready(self):
        logger.info('Waiting for database...')
        for log in self.docker_runner.run('logs', '--follow', self.container_name, _err_to_out=True, _iter='out'):
            if self.started_log in log:
                logger.info('Database ready!')
                break

    @property
    @abc.abstractmethod
    def started_log(self):
        raise NotImplementedError

    def migrate(self):
        if not self._migrated:
            logger.info('Running migrations...')
            self.docker_runner.run(
                'run',
                '--net', self.network,
                '--rm',
                *self.volumes,
                self.tested_image_name,
                self.context['TEST_MIGRATION_COMMAND'],
                _err_to_out=True,
            )
            logger.info('Migrations done!')
        self._migrated = True

    @property
    def network(self) -> str:
        return f'container:{self.container_name}'


class PostgresDatabase(Database):
    started_log = 'PostgreSQL init process complete; ready for start up.'

    def create(self):
        logger.info('Setting up database...')
        self.docker_runner.run(
            'run',
            '--restart', 'always',
            '--net', 'none',
            '--name', self.container_name,
            '--detach',
            '-e', 'POSTGRES_DB={}'.format(self.context['TEST_DATABASE_NAME']),
            self.context['TEST_DATABASE_IMAGE'],
        )


class CockroachDatabase(Database):
    started_log = 'initialized new cluster'

    def create(self):
        logger.info('Releasing cockroaches...')
        self.docker_runner.run(
            'run',
            '--restart', 'always',
            '--net', 'none',
            '--name', self.container_name,
            '--detach',
            self.context['TEST_DATABASE_IMAGE'],
            'start-single-node',
            '--insecure',
            '--host=localhost',
            '--logtostderr',
        )
        self._create_database()

    def _create_database(self):
        self.wait_until_ready()
        self.docker_runner.run(
            'run',
            '--net', self.network,
            '--rm',
            self.context['TEST_DATABASE_IMAGE'],
            'sql',
            '--insecure',
            '-e',
            'CREATE DATABASE {}'.format(self.context['TEST_DATABASE_NAME']),
        )


DATABASE_MAP = {
    'postgres': PostgresDatabase,
    'cockroach': CockroachDatabase,
}



import pathlib

from setuptools import find_packages
from setuptools import setup


def templates():
    for path in pathlib.Path('kubeyard/templates').glob('**/*'):
        if not path.is_dir():
            yield str(path.relative_to('kubeyard'))


def parse_requirements(filename, *args, **kwargs):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]


def get_long_description():
    with open('README.md') as readme_file:
        return readme_file.read()


setup(
    name='kubeyard',
    version='1.1.0',
    description='A utility to develop, test and deploy Kubernetes microservices.',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    author='Social WiFi',
    author_email='it@socialwifi.com',
    url='https://github.com/socialwifi/kubeyard',
    packages=find_packages(exclude=['tests']),
    install_requires=[str(r) for r in parse_requirements('base_requirements.txt', session=False)],
    test_suite='tests',
    entry_points={
        'console_scripts': [
            'kubeyard = kubeyard.entrypoints.kubeyard:cli_with_custom_commands',
            'sw-cli = kubeyard.entrypoints.kubeyard:cli_with_custom_commands',  # TODO: remove legacy
        ],
    },
    package_dir={'kubeyard': 'kubeyard'},
    package_data={'kubeyard': list(templates())},
    include_package_data=True,
    license='BSD',
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
)



import logging
import sys
import typing

import sh

from kubeyard.commands.devel import BaseDevelCommand

logger = logging.getLogger(__name__)


class FixCodeStyleCommand(BaseDevelCommand):
    """
    Fix code style.

    Image should have `fix_code_style` command available.

    You also need to configure volumes to apply changes to code on your machine.
    """
    custom_script_name = 'fix_code_style'

    @property
    def volumes(self) -> typing.Iterable[str]:
        if self.is_development:
            mounted_project_dir = self.cluster.get_mounted_project_dir(self.project_dir)
            for volume in self.context.get('DEV_MOUNTED_PATHS', []):
                if 'mount-in-tests' in volume and volume['mount-in-tests']['image-name'] == self.image_name:
                    host_path = str(mounted_project_dir / volume['host-path'])
                    container_path = volume['mount-in-tests']['path']
                    yield from ['-v', '{}:{}:rw'.format(host_path, container_path)]

    def run_default(self):
        try:
            self.docker_runner.run(
                'run',
                '--rm',
                *self.volumes,
                self.image,
                'fix_code_style',
                _out=sys.stdout.buffer,
                _err=sys.stdout.buffer,
            )
        except sh.ErrorReturnCode_127:
            instruction = 'You should implement `fix_code_style` command in the docker image!'
            raise NotImplementedError(instruction)



import logging
import sys

from kubeyard.commands.devel import BaseDevelCommand

logger = logging.getLogger(__name__)


class UpdateRequirementsCommand(BaseDevelCommand):
    """
    Command can update requirements using `freeze_requirements` command in container.

    Requirements: \n
        - `freeze_requirements command` available in container \n
        - volume with requirements files mounted \n

    Can be overridden in <project_dir>/scripts/update_requirements.
    """
    custom_script_name = 'update_requirements'

    def run_default(self):
        logger.info('Updating requirements for "{}"...'.format(self.image))
        self.docker_runner.run(
            'run', '--rm', '-i',
            '-u', '{}:{}'.format(self.uid, self.gid),
            '-e', 'CUSTOM_COMPILE_COMMAND="kubeyard update_requirements"',
            '-e', 'HOME=/tmp',
            *self.volumes,
            self.image,
            'bash', '-c', 'freeze_requirements',
            _out=sys.stdout.buffer,
            _err=sys.stdout.buffer,
        )
        logger.info('Requirements updated!')



import getpass
import logging
import sys

import kubepy.appliers
import sh

from cached_property import cached_property
from kubepy import appliers_options

from kubeyard import base_command
from kubeyard import kubernetes
from kubeyard import settings
from kubeyard.commands.devel import MAX_JOB_RETRIES
from kubeyard.commands.devel import BaseDevelCommand
from kubeyard.commands.devel import DockerRunner

logger = logging.getLogger(__name__)


class DeployCommand(BaseDevelCommand):
    """
    Deploys application to kubernetes.

    Next step is creating secret. Secret is named KUBE_SERVICE_NAME configured through context. its content is gathered
    either form repository in development mode or from global directory in production mode (see kubeyard help setup).
    key value pairs of this secrets are collected from files in this directory: keys are filenames and values are
    contents of these files. File secrets.yml is exception: it contains yaml encoded dictionary of additional key,
    value pairs. In genereal you should use secrets.yml for your short text secrets.

    Last step is deploying ./config/kubernetes/deploy/ using kubepy_deploy_all command
    (https://github.com/socialwifi/kubepy/). In development mode it also merges differences from
    config/development_overrides, and adds volumes to every pod configured in dev_mounted_paths in config/kubeyard.yml.
    These volumes should be mounted in selected pods using development_overrides.

    \b
    Example:
    dev_mounted_paths:
    - name: dev-volume
      host-path: docker/source

    Can be overridden in <project_dir>/sripts/deploy.
    """
    custom_script_name = 'deploy'
    context_vars = ["build_url", "aws_credentials", "gcs_service_key_file", "azure_connection_string", "bucket_name",
                    "upload_local_binary_path"]

    def __init__(self, *, build_url, gcs_service_key_file, aws_credentials, azure_connection_string, bucket_name,
                 upload_local_binary_path, **kwargs):
        super().__init__(**kwargs)
        self.build_url = build_url
        self.gcs_service_key_file = gcs_service_key_file
        self.aws_credentials = aws_credentials
        self.azure_connection_string = azure_connection_string
        self.bucket_name = bucket_name
        self.upload_local_binary_path = upload_local_binary_path

    def run_default(self):
        if self.should_deploy_statics:
            self.run_statics_deploy()
        if self.definition_directories:
            if self.dev_requirements and self.is_development:
                self.run_dev_requirements_deploy()
            self.run_kubernetes_deploy()
        if self.is_development:
            DomainConfigurator(self.context).configure()

    @property
    def should_deploy_statics(self):
        return not self.is_development and self.static_files_storage

    @cached_property
    def static_files_storage(self):
        return static_files_storage_factory(
            self.context,
            self.image,
            self.gcs_service_key_file,
            self.aws_credentials,
            self.azure_connection_string,
            self.bucket_name,
            self.upload_local_binary_path,
        )

    def run_statics_deploy(self):
        logger.info('Uploading static files...')
        self.static_files_storage.collect_and_upload()
        logger.info('Static files uploaded')

    def run_kubernetes_deploy(self):
        pod_annotations = {}
        if self.build_url is not None:
            pod_annotations['kubeyard/build-url'] = self.build_url
        options = appliers_options.Options(
            build_tag=self.tag, replace=self.is_development, host_volumes=self.host_volumes,
            max_job_retries=MAX_JOB_RETRIES, pod_annotations=pod_annotations,
        )
        kubernetes.install_secrets(self.context)
        logger.info('Applying Kubernetes definitions from YAML files...')
        kubepy.appliers.DirectoriesApplier(self.definition_directories, options).apply_all()
        logger.info('Kubernetes definitions applied')

    @property
    def definition_directories(self):
        kubernetes_dir = self.project_dir / settings.DEFAULT_KUBERNETES_DEPLOY_DIR
        overrides_dir = self.project_dir / settings.DEFAULT_KUBERNETES_DEV_DEPLOY_OVERRIDES_DIR
        definition_directories = []
        if kubernetes_dir.exists():
            definition_directories.append(kubernetes_dir)
        if self.is_development and overrides_dir.exists():
            definition_directories.append(overrides_dir)
        return definition_directories

    @property
    def host_volumes(self):
        if self.is_development:
            mounted_project_dir = self.cluster.get_mounted_project_dir(self.project_dir)
            return {
                volume['name']: mounted_project_dir / volume['host-path']
                for volume in self.context.get('DEV_MOUNTED_PATHS', [])
            }
        else:
            return {}

    def run_dev_requirements_deploy(self):
        logger.info('Checking development requirements...')
        from kubeyard.commands.dev_requirements import RequirementsDispatcher
        dispatcher = RequirementsDispatcher(self.context)
        dispatcher.dispatch_all(self.dev_requirements)
        logger.info('Development requirements are satisfied')

    @property
    def dev_requirements(self):
        return self.context.get('DEV_REQUIREMENTS')


class DomainConfigurator:
    hosts_watermark = '# The following line is added by kubeyard\n'
    host_format = '{minikube_ip}\t{domain}\n'
    hosts_filename = '/etc/hosts'

    def __init__(self, context: dict):
        self.context = context

    def configure(self):
        if self.custom_domains_to_be_configured:
            logger.info(
                f'{len(self.custom_domains_to_be_configured)} domains need to be configured: '
                f'{self.custom_domains_to_be_configured}. '
                f'Updating {self.hosts_filename}...',
            )
            self.run_update_hosts()
        else:
            logger.info('All domains already configured, no action required.')

    def run_update_hosts(self):
        for domain in self.custom_domains_to_be_configured:
            hosts_entry = self.host_format.format(minikube_ip=self.minikube_ip, domain=domain)
            sh.sudo(
                '-S',
                'tee', '--append', self.hosts_filename,
                _in=self._sudo_password + self.hosts_watermark + hosts_entry,
            )

    @cached_property
    def minikube_ip(self) -> str:
        return sh.kubectl.get.nodes(
            '-l', 'minikube.k8s.io/name=minikube',
            '-o', 'jsonpath={.items[*].status.addresses[?(@.type=="InternalIP")].address}').strip()

    @cached_property
    def _sudo_password(self):
        prompt = f"[sudo] password for {getpass.getuser()}: "
        return getpass.getpass(prompt=prompt) + "\n"

    @cached_property
    def custom_domains_to_be_configured(self) -> [str]:
        result = []
        top_level_domain = self.context['DEV_TLD']
        for domain in self.context['DEV_DOMAINS']:
            domain = f'{domain}.{top_level_domain}'
            if not self.domain_exists_in_hosts(domain):
                result.append(domain)
        return result

    def domain_exists_in_hosts(self, domain: str) -> bool:
        previous_line = ''
        with open(self.hosts_filename) as hosts_file:
            for line in hosts_file:
                if domain in line:
                    if previous_line != self.hosts_watermark:
                        logger.warning(f'Hostname: {domain} not added by kubeyard, please remove manually added entry.')
                    return True
                previous_line = line
        return False


def static_files_storage_factory(context, image, gcs_service_key_file, aws_credentials, azure_connection_string,
                                 bucket_name, local_binary_path):
    statics_directory = context.get('STATICS_DIRECTORY', '')
    collect_statics_command = context.get('COLLECT_STATICS_COMMAND', 'collect_statics_tar')
    docker_runner = DockerRunner(context)
    bucket_name = bucket_name or context.get('BUCKET_NAME')
    arguments = {
        'statics_directory': statics_directory,
        'collect_statics_command': collect_statics_command,
        'image': image,
        'docker_runner': docker_runner,
        'bucket_name': bucket_name,
        'local_binary_path': local_binary_path,
    }
    if bucket_name:
        if gcs_service_key_file:
            return GCSFilesStorage(
                **arguments,
                service_key_file=gcs_service_key_file,
            )
        elif aws_credentials:
            return S3FilesStorage(
                **arguments,
                credentials=aws_credentials,
            )
        elif azure_connection_string:
            return AzureFilesStorage(
                **arguments,
                connection_string=azure_connection_string,
            )
        else:
            raise base_command.CommandException(
                'Static file storage credential/secrets required when using bucket name!',
            )
    else:
        return None


class FilesStorage:
    def __init__(self, statics_directory, collect_statics_command, image, docker_runner, local_binary_path):
        self.statics_directory = statics_directory
        self.collect_statics_command = collect_statics_command
        self.image = image
        self.docker_runner = docker_runner
        self.local_binary_path = local_binary_path

    def collect_and_upload(self):
        statics_tar_process = self.get_statics_tar_process()
        logger.info(f'Uploading to {self.__class__.__name__}...')
        if self.local_binary_path:
            self.upload_tarred_files_with_local(statics_tar_process)
        else:
            self.upload_tarred_files(statics_tar_process)

    def get_statics_tar_process(self):
        return self.docker_runner.run(
            'run', '-i', '--rm', self.image, self.collect_statics_command, _err=sys.stdout.buffer, _piped=True)

    def upload_tarred_files(self, statics_tar_process):
        raise NotImplementedError

    def upload_tarred_files_with_local(self, statics_tar_process):
        raise NotImplementedError(
            'This storage currently does not support uploading tarred files with local binary.',
        )

    def _save_tar_to_volume(self, tar_process, volume_name):
        self.docker_runner.run_with_output(
            tar_process,
            'run', '-i', '--rm',
            '-v', '{}:/extracted/'.format(volume_name),
            'busybox:1.28.0',
            'tar', 'xf', '-', '-C', '/extracted/',
        )


class GCSFilesStorage(FilesStorage):
    cloud_sdk_image = 'google/cloud-sdk:183.0.0'

    def __init__(self, statics_directory, collect_statics_command, image, docker_runner, local_binary_path,
                 service_key_file, bucket_name):
        super().__init__(statics_directory, collect_statics_command, image, docker_runner, local_binary_path)
        self.service_key_file = service_key_file
        self.bucket_name = bucket_name

    def upload_tarred_files(self, statics_tar_process):

        with self.docker_runner.temporary_volume() as volume_name:
            self._save_tar_to_volume(statics_tar_process, volume_name)
            self.docker_runner.run_with_output(
                'run', '-i', '--rm',
                '-v', '{}:/service-account.json:ro'.format(self.service_key_file),
                '-v', '{}:/upload/:ro'.format(volume_name),
                self.cloud_sdk_image,
                'gsutil',
                '-m',
                '-o', 'Credentials:gs_service_key_file=/service-account.json',
                'rsync', '-c', '-r',
                '/upload/', 'gs://{}/{}'.format(self.bucket_name, self.statics_directory),
            )


class S3FilesStorage(FilesStorage):
    def __init__(self, statics_directory, collect_statics_command, image, docker_runner, local_binary_path,
                 credentials, bucket_name):
        super().__init__(statics_directory, collect_statics_command, image, docker_runner, local_binary_path)
        self.bucket_name = bucket_name
        if ':' in credentials:
            self.access_key, self.secret_key = credentials.split(':', 2)
        else:
            raise base_command.CommandException('AWS credentials should be in form access_key:secret_key.')

    def upload_tarred_files(self, statics_tar_process):
        upload_statics_run_command = [
            'run', '-i', '--rm',
            '-e', 'AWS_ACCESS_KEY={}'.format(self.access_key),
            '-e', 'AWS_SECRET_KEY={}'.format(self.secret_key),
            '-e', 'UPLOAD_BUCKET={}'.format(self.bucket_name),
            '-e', 'UPLOAD_PATH={}/'.format(self.statics_directory),
            'socialwifi/aws-utils:1.0.0', 'upload_tar',
        ]
        self.docker_runner.run_with_output(statics_tar_process, *upload_statics_run_command)


class AzureFilesStorage(FilesStorage):
    azure_cli_image = 'microsoft/azure-cli:2.0.61'

    def __init__(self, statics_directory, collect_statics_command, image, docker_runner, local_binary_path,
                 connection_string, bucket_name):
        super().__init__(statics_directory, collect_statics_command, image, docker_runner, local_binary_path)
        self.connection_string = connection_string
        self.bucket_name = bucket_name

    def upload_tarred_files(self, statics_tar_process):
        with self.docker_runner.temporary_volume() as volume_name:
            self._save_tar_to_volume(statics_tar_process, volume_name)
            self.docker_runner.run_with_output(
                'run', '-i', '--rm',
                '-v', '{}:/upload/:ro'.format(volume_name),
                self.azure_cli_image,
                'az',
                'storage',
                'blob',
                'upload-batch',
                '--connection-string', self.connection_string,
                '--destination', self.bucket_name,
                '--source', '/upload',
                '--destination-path', self.statics_directory,
            )

    def upload_tarred_files_with_local(self, statics_tar_process):
        statics_absolute_path = '{}/upload'.format(sh.pwd().strip())
        self._save_tar_to_volume(statics_tar_process, statics_absolute_path)
        sh.bash(
            '-c',
            " ".join([
                self.local_binary_path,
                'storage',
                'blob',
                'upload-batch',
                '--connection-string', '"{}"'.format(self.connection_string),
                '--destination', self.bucket_name,
                '--source', statics_absolute_path,
                '--destination-path', self.statics_directory,
            ]),
            _out=sys.stdout.buffer, _err=sys.stdout.buffer,
        )



import abc
import logging
import pathlib
import random
import string

import kubeyard.files_generator

from kubeyard import base_command
from kubeyard import context_factories
from kubeyard import settings

logger = logging.getLogger(__name__)


class InitCommand(base_command.BaseCommand):
    """
    Initialises empty repository. Adds basic configuration and placeholders for source code and tests.
    It can be run only in git repository. It also adds missing files in not empty repositories.
    Most commands requires repository to have files structure as provisioned by this command.
    By default it uses python application template.
    """

    def __init__(self, *, directory, init_type: 'InitType'):
        self.directory = directory
        self.init_type = init_type

    def run(self):
        logger.info("Initialising repo...")
        project_dst = pathlib.Path(self.directory)
        context = context_factories.EmptyRepoContextFactory(self.directory, self.init_type.prompted_context).get()
        template_location = f'new_repositories/{self.init_type.name}'
        kubeyard.files_generator.copy_template(template_location, project_dst, context=context)


class InitType(metaclass=abc.ABCMeta):
    name: str = NotImplemented
    prompted_context = [
        context_factories.PromptedContext(
            variable='PROJECT_NAME',
            prompt='project name',
            default=settings.DEFAULT_PROJECT_NAME_PATTERN,
        ),
        context_factories.PromptedContext(
            variable='KUBE_SERVICE_NAME',
            prompt='service name',
            default=settings.DEFAULT_KUBE_SERVICE_NAME_PATTERN,
        ),
        context_factories.PromptedContext(
            variable='KUBE_SERVICE_PORT',
            prompt='service port',
            default=settings.DEFAULT_KUBE_SERVICE_PORT,
        ),
        context_factories.PromptedContext(
            variable='DOCKER_REGISTRY_NAME',
            prompt='docker registry name',
            default=settings.DEFAULT_DOCKER_REGISTRY_NAME,
        ),
    ]


class PythonPackageInitType(InitType):
    name = 'python'


class PythonDjangoInitType(InitType):
    name = 'django'
    prompted_context = InitType.prompted_context + [
        context_factories.PromptedContext(
            variable='SECRET_KEY',
            prompt='application secret key',
            default=''.join(random.choices(string.ascii_letters + string.digits, k=50)),
        ),
    ]


class EmberInitType(InitType):
    name = 'ember'
    prompted_context = InitType.prompted_context + [
        context_factories.PromptedContext(
            variable='KUBE_LIVE_RELOAD_PORT',
            prompt='live reload development port',
            default=settings.DEFAULT_KUBE_LIVE_RELOAD_PORT,
        ),
    ]


all_templates = [
    PythonPackageInitType,
    PythonDjangoInitType,
    EmberInitType,
]



import abc
import logging
import pathlib

from cached_property import cached_property

from kubeyard import context_factories
from kubeyard import settings

logger = logging.getLogger(__name__)


class CommandException(Exception):
    pass


class BaseCommand:
    @abc.abstractmethod
    def run(self):
        raise NotImplementedError


class ContextFromClassPropertiesBuilderMetaclass(type):
    def __new__(mcs, name, bases, dct):
        dct['context_vars'] = dct.get('context_vars', [])
        for base in bases:
            if hasattr(base, 'context_vars'):
                dct['context_vars'] += base.context_vars
        obj = super().__new__(mcs, name, bases, dct)
        return obj


class InitialisedRepositoryCommand(BaseCommand, metaclass=ContextFromClassPropertiesBuilderMetaclass):
    context_vars = ['directory']

    def __init__(self, *, directory, log_level):
        super().__init__()
        self.directory = directory
        self._log_level = log_level

    def run(self):
        self.set_log_level()

    def set_log_level(self):
        logging.getLogger('kubeyard').setLevel(self.log_level)

    @property
    def log_level(self):
        return self._log_level or self.context.get('KUBEYARD_LOG_LEVEL', settings.DEFAULT_KUBEYARD_LOG_LEVEL)

    @cached_property
    def context(self):
        try:
            return context_factories.InitialisedRepoContextFactory(self.project_dir).get()
        except FileNotFoundError:
            logger.error("Invalid project root directory: {}. Exiting.".format(self.project_dir))
            exit(1)

    @cached_property
    def project_dir(self):
        return pathlib.Path(self.directory)

    @cached_property
    def custom_command_context(self):
        """
        If you want to pass any property of class, you should OVERRIDE `context_vars` class property.
        """
        context = context_factories.Context()
        context.update(self.context)
        context.update(context_factories.Context(
            {field.upper(): getattr(self, field) for field in self.context_vars},
        ))
        return context
